/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Yahoo! Mail Notifier code.
 *
 * The Initial Developer of the Original Code is
 * Michael Kraft.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// This is an object for settings and retrieving preference values.
// In addition it allows the assignment of observers and can set, retrieve or clear passwords.
// Portions of this code are based off of the preference code in the GreaseMonkey and 
// GMail Notifier Firefox extensions.

var Cc = Components.classes;
var Ci = Components.interfaces;

//	User info object
//
//  Preference object
//
var ym_prefs = {
	prefBranch : null,
	
	user : { username : null, password : null, remember_pw: false,
	         yahoo_cookie: false, overrideURL : false,
	         mailURL : "http://mail.yahoo.com", useCookie: true,
	         
	         init : function() {
		         this.username = null;
		         this.password = null;
		         this.remember_pw = false;
		         this.yahoo_cookie = false;
		         this.overrideURL = false;
		         this.mailURL = "http://mail.yahoo.com";
		         this.useCookie = true;
	         }
	},

	// pref values
	PREF_AUTO_CHECK 		: "auto-check",
	PREF_CLICK				: "click",
	PREF_DEFAULT_USER 		: "users.default", 
	PREF_DOUBLECLICK		: "doubleclick",
	PREF_DOUBLECLICK_TIME 	: "doubleclicktime",
	PREF_LOG_ENABLED		: "logenable",
	PREF_LAST_LOGIN_TIME	: "lastlogintime",
	PREF_INTERVAL 			: "interval",
	PREF_MAILURL			: "mailURL",
	PREF_MIDDLECLICK		: "middleclick",
	PREF_OVERRIDEURL		: "overrideURL",
	PREF_REMEMBER_PASSWORD	: "remember-password",
	PREF_SHOW_POPUP 		: "show-popup",
	PREF_SHOW_ALERTS		: "show-alerts",
	PREF_SOUND				: "sound",
	PREF_SOUNDFILE			: "soundfile",
	PREF_STATUSBAR			: "statusbar",
	PREF_YAHOOCOOKIE		: "yahoo-cookie", 
	PREF_USERS				: "users", 
	PREF_VERSION			: "version",
	
	// current version
	CONST_VERSION 			: "1.0.0.7",

	// alert constants
	CONST_NEVER 			:  0,
	CONST_ONCE				:  1,
	CONST_REPEAT			:  2,

	// click constants
	CONST_ONECLICK			:  0,
	CONST_TWOCLICK			:  1,
	CONST_MIDCLICK			:  2,

	// tab/window constants
	CONST_NOTHING 			:  0,
	CONST_SAMETAB 			:  1,
	CONST_NEWTAB			:  2,
	CONST_FOCUSTAB			:  3,
	CONST_NEWWIN			:  4,
	CONST_CHECKMAIL	 		:  10,
	
	// statusbar constants
	CONST_ON_NEW_MAIL 		:  1,
	CONST_ALWAYS			:  2,
	
	// how often should extension relogin to Yahoo
	CONST_RELOGIN_TIME 		: 12,
	
	// password manager URL
	CONST_PASSURL 			: "chrome://ym_notifier",
	CONST_REALM 			: "yahoo-password",

	//
	// Returns a object to the extension's preference branch
	//
	getPrefBranch : function(){
		if (!this.prefBranch){ 
		this.prefBranch = Cc['@mozilla.org/preferences-service;1'].getService(Ci.nsIPrefService).getBranch("extensions.ym_notifier.");
		}

		return this.prefBranch;
	},

	//
	// Delete user login information
	//
	deleteUser : function(userName) {
		var pref = this.getPrefBranch();
		var userBranch = this.PREF_USERS + "." + userName + ".";

		// remove user from users list
		var newUsers = new Array();
		var users = this.getPref(this.PREF_USERS).split(" ");
		for (var i=0; i < users.length; i++) {
			if (users[i] != userName) {
				newUsers.unshift(users[i]);
			}
		}
		this.setPref(this.PREF_USERS, newUsers.join(" "));

		if (userName == this.getPref(this.PREF_DEFAULT_USER)) {
			if (newUsers[0]) this.setPref(this.PREF_DEFAULT_USER, newUsers[0]);
			else pref.clearUserPref(this.PREF_DEFAULT_USER);
		}

		if (this.exists(userBranch + this.PREF_REMEMBER_PASSWORD)) {
			this.clearPassword(userName);
			this.deletePref(userBranch + this.PREF_REMEMBER_PASSWORD);
			this.deletePref(userBranch + this.PREF_YAHOOCOOKIE);
			this.deletePref(userBranch + this.PREF_LAST_LOGIN_TIME);
			this.deletePref(userBranch + this.PREF_OVERRIDEURL);
			this.deletePref(userBranch + this.PREF_MAILURL);
		}
	},

	//
	// Save user login information
	//
	saveUser : function(userName, password, remember_password, overrideURL, mailURL, make_default) {
		var userBranch = this.PREF_USERS + "." + userName + ".";

		// initialize user preferences if not already set
		this.initPref(userBranch + this.PREF_YAHOOCOOKIE, "");
		this.initPref(userBranch + this.PREF_LAST_LOGIN_TIME, 0);

		if (remember_password && (password != "")) this.setPassword(password, userName);
		else this.clearPassword(userName);
		this.setPref(userBranch + this.PREF_REMEMBER_PASSWORD, remember_password);
		this.setPref(userBranch + this.PREF_OVERRIDEURL, overrideURL);
		this.setPref(userBranch + this.PREF_MAILURL, mailURL);

		if (make_default) this.setPref(this.PREF_DEFAULT_USER, userName);

		// add user to users list
		var users = this.getPref(this.PREF_USERS)
		if (users) users = this.getPref(this.PREF_USERS).split(" ");
		else users = new Array();
		var foundUser = false;
		for (var i=0; i < users.length; i++) {
			if (users[i] == userName) {
				foundUser = true;
				break;
			}
		}

		if (!foundUser) {
			users.unshift(userName);
			users = users.sort();
			this.setPref(this.PREF_USERS, users.join(" "));
		}
	},

	//
	// Save user cookie
	//
	saveUserCookie : function(cookie, userName) {
		// if no username specified retrieve default user info
		if (!userName || userName=="undefined") userName = this.getPref(this.PREF_DEFAULT_USER);

		if (userName) {
			var cookiePref = this.PREF_USERS + "." + userName + "." + this.PREF_YAHOOCOOKIE;
			var lastloginPref = this.PREF_USERS + "." + userName + "." + this.PREF_LAST_LOGIN_TIME;

			// get current date/time
			var now = new Date();

			this.setPref(cookiePref, cookie);
			if (cookie) {
		 		// store last login in hours since 1970 (
		 		this.setPref(lastloginPref, parseInt(now.getTime()/3600000));
	 		}
			else {
		 		this.setPref(lastloginPref, 0);
	 		}
		}
	},

	//
	// Save user mailURL
	//
	saveUserMailURL : function(mailURL, userName) {
		// if no username specified retrieve default user info
		if (!userName || userName=="undefined") userName = this.getPref(this.PREF_DEFAULT_USER);
		if (userName) {
			var mailURLPref = this.PREF_USERS + "." + userName + "." + this.PREF_MAILURL;

			this.setPref(mailURLPref, mailURL);
		}
	},

	//
	// Retrieve user login information
	//
	loadUser : function(userName, loadPw) {
		// Initialize user data
		this.user.init();
		
		// if no username specified retrieve default user info
		if (!userName || userName=="undefined") userName = this.getPref(this.PREF_DEFAULT_USER);

		var userBranch = this.PREF_USERS + "." + userName + ".";

		// if username not stored (remember password pref missing) return null;
		try {
			// no username means no info
			if (userName) {
				this.user.userName = userName;
				if (loadPw != false) this.user.password = this.getPassword(userName);
				this.user.remember_pw = this.getPref(userBranch + this.PREF_REMEMBER_PASSWORD);
			}
		}
		catch (ex) {
			this.user.init();
		}

		if (this.user.userName) {
	
			try { 
				this.user.overrideURL = this.getPref(userBranch + this.PREF_OVERRIDEURL);
			} catch (ex) { 
				this.initPref(userBranch + this.PREF_OVERRIDEURL, false);
				this.user.overrideURL = false;
			}

			try { 
				this.user.mailURL = this.getPref(userBranch + this.PREF_MAILURL);
			} catch (ex) { 
				this.initPref(userBranch + this.PREF_MAILURL, "http://mail.yahoo.com");
				this.user.mailURL = "http://mail.yahoo.com";
			}
	
			try { 
				this.user.cookie = this.getPref(userBranch + this.PREF_YAHOOCOOKIE);
				var last_login_time = this.getPref(userBranch + this.PREF_LAST_LOGIN_TIME);
				var now = new Date();

				// if over 12 hours since last login clear cookie to force normal login
				if ((parseInt(now.getTime()/3600000) - last_login_time) > this.CONST_RELOGIN_TIME)
					this.user.useCookie = false;

			} catch (ex) { 
				this.initPref(userBranch + this.PREF_YAHOOCOOKIE, "");
				this.initPref(userBranch + this.PREF_LAST_LOGIN_TIME, 0);
				this.user.cookie = "";
				this.user.useCookie = false;
			}
		}
	},

	//
	// Whether a preference exists
	//
	exists : function(prefName) {
		var pref = this.getPrefBranch();
		return pref.getPrefType(prefName) != pref.PREF_INVALID;
	},

	//
	// Returns the named preference, or defaultValue if it does not exist
	//
	getPref : function(prefName) {
		var pref = this.getPrefBranch();

		var prefType = pref.getPrefType(prefName);

		// underlying preferences object throws an exception if pref doesn't exist
		if (prefType == pref.PREF_INVALID) {
			throw new Error("Preference is not set: " + prefName);
		}
	
		switch (prefType) {
		case pref.PREF_STRING:
			// handle unicode values
			return pref.getComplexValue(prefName,Ci.nsISupportsString).data
			break;
		case pref.PREF_BOOL:
			return pref.getBoolPref(prefName);
			break;
		case pref.PREF_INT:
			return pref.getIntPref(prefName);
			break;
		}

		return null;
	},

	//
	// Sets the named preference to the specified value. values must be strings, booleans, or integers.
	//
	setPref : function(prefName, value) {
		var prefType = typeof(value);
	
		switch (prefType) {
		case "string":
		case "boolean":
			break;
		case "number":
			if (value % 1 != 0) {
			throw new Error("Cannot set preference to non integral number");
			}
			break;
		default:
			throw new Error("Cannot set preference with datatype: " + prefType);
		}
	
		// underlying preferences object throws an exception if new pref has a 
		// different type than old one. i think we should not do this, so delete
		// old pref first if this is the case.
		//
		if (this.exists(prefName) && prefType != typeof(this.getPref(prefName))) {
			this.deletePref(prefName);
		}

		// set new value using correct method
		switch (prefType) {
			case "string":
				// handle unicode values
				var str = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
				str.data = value;
				this.getPrefBranch().setComplexValue(prefName,Ci.nsISupportsString, str);
				break;
			case "boolean":
				this.getPrefBranch().setBoolPref(prefName, value);
				break;
			case "number":
				this.getPrefBranch().setIntPref(prefName, Math.floor(value));
				break;
		}
	},
	
	//
	// Deletes the named preference or subtree
	//
	deletePref : function(prefName) {
		this.getPrefBranch().deleteBranch(prefName);
	},

	//
	// Initializes the default preference values. this should not be required since
	// the default values should be initiated by the defaults file, but keep this
	// here to be safe.
	//
	initPrefs : function() {
		var arePrefsUpToDate = this.exists(this.PREF_VERSION);
		if (!arePrefsUpToDate)
		{
			// set prefs if they don't exist
			this.initPref(this.PREF_AUTO_CHECK, true);
			this.initPref(this.PREF_CLICK, this.CONST_NOTHING);
			this.initPref(this.PREF_DEFAULT_USER, "");
			this.initPref(this.PREF_DOUBLECLICK, this.CONST_SAMETAB);
			this.initPref(this.PREF_DOUBLECLICK_TIME, 400);
			this.initPref(this.PREF_INTERVAL, 5);
			this.initPref(this.PREF_MIDDLECLICK, this.CONST_FOCUSTAB);
			this.initPref(this.PREF_SHOW_POPUP, this.CONST_NEVER);
			this.initPref(this.PREF_SHOW_ALERTS, this.CONST_NEVER);
			this.initPref(this.PREF_SOUND, this.CONST_NEVER);
			this.initPref(this.PREF_SOUNDFILE, "");
			this.initPref(this.PREF_STATUSBAR, this.CONST_ALWAYS);
			this.initPref(this.PREF_LOG_ENABLED, false);
			this.initPref(this.PREF_USERS, "");
		}

		// update preferences that should change with this version
		if (!arePrefsUpToDate || (this.getPref(this.PREF_VERSION) != this.CONST_VERSION))
		{
			this.setPref(this.PREF_VERSION, this.CONST_VERSION);
		}
	},

	//
	// If the preference doesn't already exist, set it
	//
	initPref : function(aPrefName, aDefaultValue) {
		var prefExists = this.exists(aPrefName);
		if (!prefExists) this.setPref(aPrefName, aDefaultValue);
	},


	//
	// Add an observer to call the observer aObserver when preference aPrefName changes
	//
	addObserver : function(aPrefName, aObserver) {
		var myPrefs = this.getPrefBranch();
		var prefBranchInternal = myPrefs.QueryInterface(Ci.nsIPrefBranchInternal);

		if (prefBranchInternal) prefBranchInternal.addObserver(aPrefName, aObserver, false);
	},

	//
	// Remove observer that calls the observer aObserver when preference aPrefName changes
	//
	removeObserver : function(aPrefName, aObserver) {
		var myPrefs = this.getPrefBranch();
		var prefBranchInternal = myPrefs.QueryInterface(Ci.nsIPrefBranchInternal);

		if (prefBranchInternal) prefBranchInternal.removeObserver(aPrefName, aObserver);
	},

	
	//
	//	The following functions have nothing to do with the preference interface but
	//	since they are preference related they are included here.
	//

	//
	// Get the password associated with the given username.  If username not
	// passed, retrieve password associated with stored username.
	//
	getPassword : function(username) {
		// if no username specified retrieve default user info
		if (username == null) username = this.getPref(this.PREF_DEFAULT_USER);

		// no username means no password
		if (username == "") return '';

		// Get Password Manager (does not exist in Firefox 3.0)
		var CC_passwordManager = Cc["@mozilla.org/passwordmanager;1"]
	 
		if (CC_passwordManager != null) {
			
			// Used by all except Firefox 3.0
			var passwordManager = CC_passwordManager.getService(Ci.nsIPasswordManagerInternal);
		
			var host = {value:""};
			var user =	{value:""};
			var pass = {value:""}; 
	 
			try {
				passwordManager.findPasswordEntry(this.CONST_PASSURL, username, "", host, user, pass);
				return (pass.value);
			}
			catch(ex) {
				dump(ex);
			}
		} 
		// Firefox 3.0 browser
		else {
			try {

				// Get Login Manager (Firefox 3.0)
				var passwordManager = Cc["@mozilla.org/login-manager;1"].getService(Ci.nsILoginManager);
		
				// Find users for this extension 
				var logins = passwordManager.findLogins({}, this.CONST_PASSURL, null, this.CONST_REALM);
		
				// If no passwords try to change password format
				if (!logins.length) logins = this.migratePasswords();
		
				for (var i = 0; i < logins.length; i++) {
					if (logins[i].username == username) {
						return logins[i].password;
						break;
					}
				}
			}
			catch(ex) {
				dump(ex);
			}
		}
		return '';
	},

	//
	// Store and associated a password with the given username.  If username not
	// passed, store and associate password with stored username.
	//
	setPassword : function(password, username) {
		// if no username specified retrieve default user info
		if (username == null) username = this.getPref(this.PREF_DEFAULT_USER);

		// set the password
		if (username != "") this.setClearPassword(username, password, true);
	},

	//
	// Clear the password associated with the given username.  If username not
	// passed, clear password associated with stored username.
	//
	clearPassword : function(username) {
		// if no username specified retrieve default user info
		if (username == null) username = this.getPref(this.PREF_DEFAULT_USER);

		// clear the password
		this.setClearPassword(username, "", false);
	},

	// 
	// Common code to set/clear password
	//
	setClearPassword : function(username, password, set) {
		// Get Password Manager (does not exist in Firefox 3.0)
		var CC_passwordManager = Cc["@mozilla.org/passwordmanager;1"]
	 
		if (CC_passwordManager != null) {
			// Used by all except Firefox 3.0
			var passwordManager = CC_passwordManager.getService(Ci.nsIPasswordManager);
		
			// try to remove the current password, if it's there
			try {
				passwordManager.removeUser(this.CONST_PASSURL, username);
			} 
			catch (ex) {
				dump(ex);
			}
			
			if (set) passwordManager.addUser(this.CONST_PASSURL, username, password);
		}
		// Firefox 3.0 browser
		else {
			// Get Login Manager (Firefox 3.0)
			var passwordManager = Cc["@mozilla.org/login-manager;1"].getService(Ci.nsILoginManager);
							
			var oldLogin = Cc["@mozilla.org/login-manager/loginInfo;1"].createInstance(Ci.nsILoginInfo);
					 
			var newLogin = Cc["@mozilla.org/login-manager/loginInfo;1"].createInstance(Ci.nsILoginInfo);

			oldLogin.init(this.CONST_PASSURL, null, this.CONST_REALM, username, this.getPassword(username), "", "");
			newLogin.init(this.CONST_PASSURL, null, this.CONST_REALM, username, password, "", "");
	
			// try to remove the current password, if it's there
			if (oldLogin.password != "") {
	 			passwordManager.removeLogin(oldLogin);
			} 
	
			if (set) passwordManager.addLogin(newLogin);
		}
	},

	// 
	// Change password format for Firefox 3.0
	//
	migratePasswords : function() {
		// Get Login Manager (Firefox 3.0)
		var passwordManager = Cc["@mozilla.org/login-manager;1"].getService(Ci.nsILoginManager);

		var nsLoginInfo = new Components.Constructor("@mozilla.org/login-manager/loginInfo;1", Ci.nsILoginInfo);

		// Find users for this extension 
		var logins = passwordManager.findLogins({}, this.CONST_PASSURL, "/", null);
		
		for (var i = 0; i < logins.length; i++) {
			try {
				newLogin = new nsLoginInfo();
				newLogin.init(this.CONST_PASSURL, null, this.CONST_REALM, logins[i].username, logins[i].password, "", "");
				passwordManager.modifyLogin(logins[i], newLogin);
			}
			catch (ex) {}
		}
		return logins;
	} 
}

// Make sure prefs are up to date
ym_prefs.initPrefs();